#!/usr/bin/perl -w

###########################################################################
#
# writetranslationmedium.pl
#
# A component of the Greenstone digital library software
# from the New Zealand Digital Library Project at the 
# University of Waikato, New Zealand.
#
# Copyright (C) 1999 New Zealand Digital Library Project
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
###########################################################################


BEGIN {
    die "GSDLHOME not set\n" unless defined $ENV{'GSDLHOME'};
    unshift (@INC, "$ENV{'GSDLHOME'}/perllib");
}


###########################################################################
#
# CODE SYNOPSIS
#
# This script writes to some medium (currently HTML or Excel) the chunks
#   needing translation or updating, and spaces for the translations to be
#   entered. It:
#
# - Parses the script arguments.
# - Checks that the source and target languages are specified in the
#     etc/main.cfg file.
# - Reads the translation files from the translator configuration file,
#     and check that the one specified is defined.
# - Checks that the translation has previously been initialised.
# - Reads the translation file data from the config file.
# - Determines the chunks needing writing to file:
#   - All the chunks (if 'show_all_translation' option is specified), or
#   - Those chunks needing translation or updating (otherwise)
# - Writes these chunks to the specified file
#
###########################################################################


use translator;
use util;
use GDBM_File;
use CGI;


my $gsdldir = "$ENV{'GSDLHOME'}";

my $option_show_all_translations = 0;
my $option_show_complete_translations = 0;


sub main
{
    # Get the code of the source (base) language
    local $sourceabbr = shift(@_);
    # Get the code of the target language
    local $targetabbr = shift(@_);
    # Get the name of the file to translate
    local $transfile = shift(@_);
    # Get the type of media to output to
    local $outputmedia = shift(@_);

    # Read any extra arguments
    local $optionalarg = shift(@_);
    while ($optionalarg) {
	if ($optionalarg eq "-show_all_translations") { $option_show_all_translations = 1; }
	if ($optionalarg eq "-show_complete_translations") { $option_show_complete_translations = 1; }
	$optionalarg = shift(@_);
    }

    # Log any configuration errors to the general translator log
    local $translatordir = &translator::get_translator_dir();
    local $translatorlogfile = &util::filename_cat($translatordir, "translator.log");
    if (!open(LOG, ">>$translatorlogfile")) {
	die "Error: Could not append to log file $translatorlogfile.\n";
    }

    # Log the command that launched this script
    print LOG "\nCommand: $0 @ARGV\n";

    # Check that the necessary arguments were supplied
    if (!$sourceabbr || !$targetabbr || !$transfile || !$outputmedia) {
	print LOG &get_usage_description();
	print LOG "Error: Required argument missing.\n";
	die "\n";
    }

    # Check that the languages are specified in the etc/main.cfg file
    local %langhash = &translator::read_languages_from_main_cfg_file();
    local $sourcelang = $langhash{$sourceabbr};
    if (!$sourcelang) {
	print LOG "Error: Language \"$sourceabbr\" is not declared in etc/main.cfg.\n";
	die "\n";
    }
    local $targetlang = $langhash{$targetabbr};
    if (!$targetlang) {
	print LOG "Error: Language \"$targetabbr\" is not declared in etc/main.cfg.\n";
	die "\n";
    }

    # Check that the file to translate is defined in the translator config file
    local %files = &translator::read_translation_files_from_config_file();
    if (!$files{$transfile}) {
	print LOG "Error: File \"$transfile\" is not defined in etc/translator.cfg.\n";
	die "\n";
    }

    # Make sure this is a continuation of a previously initialised translation process
    local $translatordir = &translator::get_translator_dir();
    local $languagedir = &util::filename_cat($translatordir, "$sourceabbr-$targetabbr");
    if (!-e $languagedir) {
	print LOG "Error: Directory $languagedir does not exist (translation not initialised).\n";
	die "\n";
    }

    # Make sure a translation database exists for this file
    local $translationdir = &util::filename_cat($languagedir, $transfile);
    local $translationdb = &util::filename_cat($translationdir, "translation.db");
    if (!-e $translationdb) {
	print LOG "Error: No translation database exists for $transfile.\n";
	die "\n";
    }

    # Read the details of the file to translate from the translator config file
    local ($sourcefile, $targetfile, $filetype) =
	&translator::read_file_data_from_config_file($sourceabbr, $targetabbr, $transfile);

    # Log all the rest of the output to the log for this translation process
    close(LOG);
    local $translationlogfile = &util::filename_cat($translationdir, "translation.log");
    if (!open(LOG, ">>$translationlogfile")) {
	die "Error: Could not append to log file $translationlogfile.\n";
    }

    # Log the command that launched this script
    print LOG "\nCommand: $0 @ARGV\n";


    # -----------------------------------------
    #   Refresh the translation database
    # -----------------------------------------

    &translator::refresh_translation($translationdir, $sourcefile, $filetype);


    # -----------------------------------------
    #   Read the source language file
    # -----------------------------------------

    # Parse the source language file, building a mapping from chunk key to chunk text
    local $sourcepath = &translator::get_translation_file_path($translationdir, $sourcefile);
    local %sourcehash = &translator::process_file($sourcepath, $filetype);
    print LOG "Number of source chunks: " . scalar(keys(%sourcehash)) . "\n";

    # Open the translation database for this file
    local %translationdata = ();
    tie(%translationdata, "GDBM_File", $translationdb, GDBM_READER, 0666);

    # Chunks to write are...
    local %translationstowrite = ();

    # ...all of them (if 'show_all_translations' has been specified)...
    if ($option_show_all_translations) {
  	foreach $key (keys(%translationdata)) {
  	    if ($key !~ /\*$/) {
  		$translationstowrite{$key}[0] = $sourcehash{$key};
  		$translationstowrite{$key}[1] = $translationdata{$key} || "";
  	    }
  	}
    }

    elsif ($option_show_complete_translations) {
  	foreach $key (keys(%translationdata)) {
  	    if ($key !~ /\*$/ && $translationdata{$key} ne "") {
  		$translationstowrite{$key}[0] = $sourcehash{$key};
  		$translationstowrite{$key}[1] = $translationdata{$key};
  	    }
  	}
    }

    # ...or (otherwise)...
    else {
  	foreach $key (keys(%translationdata)) {
  	    # ...those marked as needing updating...
  	    if ($key =~ /\*$/) {
  		chop($key);
  		$translationstowrite{$key}[0] = $sourcehash{$key};
  		$translationstowrite{$key}[1] = $translationdata{$key} || "";
  		print LOG "Chunk with key $key needs updating.\n";
  	    }

  	    # ...and those in the source file where no translation exists in the database
  	    elsif ($translationdata{$key} eq "" && $sourcehash{$key} ne "") {
  		$translationstowrite{$key}[0] = $sourcehash{$key};
  		$translationstowrite{$key}[1] = "";
  		print LOG "No translation exists for chunk with key $key.\n";
  	    }
  	}
    }

    # print LOG "Chunks to write:\n";  
    # foreach $key (keys(%translationstowrite)) {
    #     print LOG "$key\n";
    #     print LOG "Source:\n" . $translationstowrite{$key}[0] . "\n";
    #     print LOG "Target:\n" . $translationstowrite{$key}[1] . "\n";
    # }

    # Output the translations needing writing, in the required format
    if ($outputmedia eq "html") {
	# Output as a series of HTML pages
	&write_html_pages($sourceabbr, $sourcelang, $targetabbr, $targetlang, $transfile,
			  $translationdir, $sourcefile, %translationstowrite);
    }
    elsif ($outputmedia eq "excel") {
	# Output as a XML Microsoft Excel document
	&write_excel_file($sourceabbr, $sourcelang, $targetabbr, $targetlang, $transfile,
			  $translationdir, $sourcefile, %translationstowrite);
    }
    elsif ($outputmedia eq "text") {
	# Just output the text strings needing translation
	&write_text_file($sourceabbr, $sourcelang, $targetabbr, $targetlang, $transfile,
			 $translationdir, $sourcefile, %translationstowrite);
    }
    else {
	print LOG "Error: Unknown output media \"$outputmedia\".\n";
    }

    # Clean up
    untie(%translationdata);
}


sub get_usage_description
{
    return "Usage: writetranslationmedium.pl <sourceabbr> <targetabbr> <transfile> html|excel [-show_all_translations]\n";
}


sub write_html_pages
{
    my ($sourceabbr, $sourcelang, $targetabbr, $targetlang, $transfile,
	$translationdir, $sourcefile, %translationstowrite) = @_;

    # Calculate the number of pages to translate
    local $pagelimit = 15;
    local $numchunks = scalar(keys(%translationstowrite));
    local $numpages = int($numchunks / $pagelimit) + (($numchunks % $pagelimit) > 0);

    print LOG "Number of chunks to translate: " . $numchunks . "\n";
    print LOG "Number of pages to translate: " . $numpages . "\n";

    local $pagenum = 0;
    local %pagehash = ();

    foreach $key (sort(keys(%translationstowrite))) {
	$pagehash{$key} = $translationstowrite{$key};

	# If the page limit is reached then write out a page
	if (scalar(keys(%pagehash)) == $pagelimit) {
	    $pagenum++;
	    &write_html_page($sourceabbr, $sourcelang, $targetabbr, $targetlang, $transfile,
			     $translationdir, $sourcefile, $pagenum, $numpages, %pagehash);

	    # Update variables
	    %pagehash = ();
	}
    }

    # Write out a page containing any remaining chunks
    if (scalar(keys(%pagehash)) != 0) {
	$pagenum++;
	&write_html_page($sourceabbr, $sourcelang, $targetabbr, $targetlang, $transfile,
			 $translationdir, $sourcefile, $pagenum, $numpages, %pagehash);
    }

    # Write a thankyou page for the language translator once the translation is complete
    $pagenum++;
    local $thankyoufile = &util::filename_cat($translationdir, "$pagenum.trans");
    if (!open(THANKYOU_FILE, ">$thankyoufile")) {
	print LOG "Error: Could not write $thankyoufile.\n";
	die "\n";
    }
    print THANKYOU_FILE "<center> _textthanks_ $targetlang _texttrans_ <p></center>\n";
    close(THANKYOU_FILE);
}


sub write_html_page
{
    my ($sourceabbr, $sourcelang, $targetabbr, $targetlang, $transfile,
	$translationdir, $sourcefile, $pagenum, $numpages, %pagehash) = @_;

    # Open a file for the HTML code of this page
    local $transpagefile = &util::filename_cat($translationdir, $pagenum . ".trans");
    if (!open(HTMLFILE, ">$transpagefile")) {
	print LOG "Error: Could not write $transpagefile.\n";
	die "\n";
    }

    # Common Gateway Interface for writing the stuff to the web
    local $cgiobj = new CGI;
    print HTMLFILE ($cgiobj->hidden(-name=>"slng", -default=>"$sourceabbr") . "\n",
		    $cgiobj->hidden(-name=>"tlng", -default=>"$targetabbr") . "\n",
		    $cgiobj->hidden(-name=>"file", -default=>"$transfile") . "\n");

    # Write "Page x of y"
    print HTMLFILE ("<center>\n",
		    "<strong>_textpage_ $pagenum _textof_ $numpages</strong>\n",
		    "</center>\n\n");

    # Write table header showing the source and target language
    print HTMLFILE ("<table cellpadding=\"10\">\n",
		    "<tr>\n",
		    "<td><center><strong>" . uc($sourcelang) . "</strong></center></td>\n",
		    "<td><center><strong>" . uc($targetlang) . "</strong></center></td>\n",
		    "</tr>\n");

    # Write a row for each of the keys needing translation
    foreach $key (sort(keys(%pagehash))) {
	local $sourcekey = "source::" . $key;
	local $targetkey = "target::" . $key;

	# Format the chunk text strings, depending on the file type
	local $sourcetext = $pagehash{$key}[0];
	local $targettext = $pagehash{$key}[1];

	# For HTML, we need to escape all the backslashes and all the tags
	$sourcetext =~ s/\\/\\\\/g;
	$sourcetext =~ s/</\&lt\;/g;
	$sourcetext =~ s/>/\&gt\;/g;
	$targettext =~ s/\\/\\\\/g;
	$targettext =~ s/</\&lt\;/g;
	$targettext =~ s/>/\&gt\;/g;

	# For display by Greenstone, we also need to escape all the underscores
	$sourcekey =~ s/_/\\_/g;
	$sourcetext =~ s/_/\\_/g;
	$targetkey =~ s/_/\\_/g;
	$targettext =~ s/_/\\_/g;

	# Write a text area containing the source text (read-only)
	local $sourcetextbox = "<textarea name=\"$sourcekey\" rows=\"3\" cols=\"50\" readonly=\"1\">";
	$sourcetextbox .= "$sourcetext</textarea>";

	# Write a text area containing the target text
	local $targettextbox = "<textarea name=\"$targetkey\" rows=\"3\" cols=\"50\">";
	$targettextbox .= "$targettext</textarea>";

	print HTMLFILE ("<tr>\n",
			"<td align=center>" . $sourcetextbox . "</td>\n",
			"<td align=center>" . $targettextbox . "</td>\n",
			"</tr>\n");
    }

    # Add "Next" and "Reset" buttons
    print HTMLFILE ("<tr>\n",
		    "<td></td>\n",
		    "<td align=right>" . $cgiobj->submit($pagenum, "_textnext_") . "<br>",
		    "_textnextwarning_" . $cgiobj->reset("_textresetform_"),
		    "</td>\n",
		    "</tr>\n",
		    "</table>\n\n");

    close(HTMLFILE);
}


sub write_text_file
{
    my ($sourceabbr, $sourcelang, $targetabbr, $targetlang, $transfile,
	$translationdir, $sourcefile, %translationstowrite) = @_;

    # Calculate the number of chunks to translate
    local $numchunks = scalar(keys(%translationstowrite));
    print LOG "Number of chunks to translate: " . $numchunks . "\n";

    # Open a file for the Excel code of this page
    local $excelfile = &util::filename_cat($translationdir, "translate.txt");
    if (!open(EXCELFILE, ">$excelfile")) {
	print LOG "Error: Could not write $excelfile.\n";
	die "\n";
    }

    local $targetfile = &util::filename_cat($translationdir, "target.txt");
    if (!open(TARGETFILE, ">$targetfile")) {
	print LOG "Error: Could not write $targetfile.\n";
	die "\n";
    }

    # Write a row for each of the keys needing translation
    foreach $key (sort(keys(%translationstowrite))) {
	# Format the source text strings, and display in a new row
	local $sourcetext = $translationstowrite{$key}[0];
	$sourcetext =~ s/</\&lt\;/g;
	$sourcetext =~ s/>/\&gt\;/g;
	print EXCELFILE "$sourcetext\n";

	# Format the source text strings, and display in a new row
	local $targettext = $translationstowrite{$key}[1];
	$targettext =~ s/</\&lt\;/g;
	$targettext =~ s/>/\&gt\;/g;
	print TARGETFILE "$targettext\n";	
    }

    close(EXCELFILE);
    close(TARGETFILE);
}


sub write_excel_file
{
    my ($sourceabbr, $sourcelang, $targetabbr, $targetlang, $transfile,
	$translationdir, $sourcefile, %translationstowrite) = @_;

    # Calculate the number of chunks to translate
    local $numchunks = scalar(keys(%translationstowrite));
    print LOG "Number of chunks to translate: " . $numchunks . "\n";

    # Open a file for the Excel code of this page
    local $excelfile = &util::filename_cat($translationdir, "excel.xml");
    if (!open(EXCELFILE, ">$excelfile")) {
	print LOG "Error: Could not write $excelfile.\n";
	die "\n";
    }

    print EXCELFILE
	("<?xml version=\"1.0\" ?>\n",
	 "<Workbook xmlns=\"urn:schemas-microsoft-com:office:spreadsheet\"\n",
	 "  xmlns:o=\"urn:schemas-microsoft-com:office:office\"\n",
	 "  xmlns:x=\"urn:schemas-microsoft-com:office:excel\"\n",
	 "  xmlns:ss=\"urn:schemas-microsoft-com:office:spreadsheet\"\n",
	 "  xmlns:html=\"http://www.w3.org/TR/REC-html40\">\n",
	 "  <DocumentProperties xmlns=\"urn:schemas-microsoft-com:office:office\">\n",
	 "    <Author>Greenstone Translator Interface</Author>\n",
	 "    <LastAuthor>Greenstone Translator Interface</LastAuthor>\n",
	 "    <Created>2003-05-05T01:30:36Z</Created>\n",
	 "    <Company>New Zealand Digital Library Project</Company>\n",
	 "    <Version>10.2625</Version>\n",
	 "  </DocumentProperties>\n",
	 "  <OfficeDocumentSettings xmlns=\"urn:schemas-microsoft-com:office:office\">\n",
	 "    <DownloadComponents/>\n",
	 "    <LocationOfComponents HRef=\"blah\"/>\n",  # !! TO DO !!
	 "  </OfficeDocumentSettings>\n",
	 "  <Styles>\n",
	 "    <Style ss:ID=\"Default\" ss:Name=\"Normal\">\n",
	 "      <Alignment ss:Vertical=\"Top\"/>\n",
	 "      <Borders/>\n",
	 "      <Font/>\n",
	 "      <Interior/>\n",
	 "      <NumberFormat/>\n",
	 "      <Protection/>\n",
	 "    </Style>\n",
	 "    <Style ss:ID=\"Heading\">\n",
         "      <Font x:Family=\"Swiss\" ss:Bold=\"1\"/>\n",
         "    </Style>\n",
	 "    <Style ss:ID=\"TextCell\">\n",
	 "      <Alignment ss:Vertical=\"Bottom\" ss:WrapText=\"1\"/>\n",
	 "      <Font ss:FontName=\"Arial Unicode MS\"/>\n",
	 "    </Style>\n",
	 "  </Styles>\n",
	 "  <Worksheet ss:Name=\"Greenstone Translation\">\n",
	 "    <Table>\n",
	 "      <Column ss:Index=\"1\" ss:AutoFitWidth=\"0\" ss:Width=\"250\"/>\n",
         "      <Column ss:Index=\"2\" ss:AutoFitWidth=\"1\" ss:Width=\"1000\"/>\n",
	 "      <Row>\n",
	 "        <Cell ss:StyleID=\"Heading\"><Data ss:Type=\"String\">Key</Data></Cell>\n",
	 "        <Cell ss:StyleID=\"Heading\"><Data ss:Type=\"String\">Text</Data></Cell>\n",
	 "      </Row>\n");

    # Write a row for each of the keys needing translation
    foreach $key (sort(keys(%translationstowrite))) {
	local $sourcekey = "source::" . $key;
	local $targetkey = "target::" . $key;
	# local $sourcekey = $transfile . "::" . "source::" . $key;  # For UNESCO stuff
	# local $targetkey = $transfile . "::" . "target::" . $key;  # For UNESCO stuff

	print EXCELFILE "      <Row/>\n";

	# Format the source text strings, and display in a new row
	local $sourcetext = $translationstowrite{$key}[0];
	$sourcetext =~ s/</\&lt\;/g;
	$sourcetext =~ s/>/\&gt\;/g;
	$sourcetext =~ s/\n/&\#10\;\n/g;
	print EXCELFILE ("      <Row>\n",
			 "        <Cell>\n",
			 "          <Data ss:Type=\"String\">$sourcekey</Data>\n",
			 "        </Cell>\n",
			 "        <Cell ss:StyleID=\"TextCell\">\n",
			 "          <Data ss:Type=\"String\">$sourcetext</Data>\n",
			 "        </Cell>\n",
			 "      </Row>\n");

	# Format the target text strings, and display in a new row
	local $targettext = $translationstowrite{$key}[1];
	$targettext =~ s/</\&lt\;/g;
	$targettext =~ s/>/\&gt\;/g;
	$targettext =~ s/\n/&\#10\;\n/g;
	print EXCELFILE ("      <Row>\n",
			 "        <Cell>\n",
			 "          <Data ss:Type=\"String\">$targetkey</Data>\n",
			 "        </Cell>\n",
			 "        <Cell ss:StyleID=\"TextCell\">\n",
			 "          <Data ss:Type=\"String\">$targettext</Data>\n",
			 "        </Cell>\n",
			 "      </Row>\n");
    }

    print EXCELFILE ("    </Table>\n",
		     "  </Worksheet>\n",
		     "</Workbook>\n");

    close(EXCELFILE);
}


&main(@ARGV);
